import pandas as pd


def _get_variable_usage(output, x):
    output = output.split("\n")
    # get the attribute usage section of the model output
    start_vars = [i for i, c in enumerate(output) if '\tAttribute usage' in c]
    # if not present raise an error
    if len(start_vars) != 1:
        raise ValueError("cannot find attribute usage data")
    output = output[start_vars[0]:(len(output))-2]
    output = [c.replace("\t", "") for c in output]
    has_pct = [i for i, c in enumerate(output) if "%" in c]
    if len(has_pct) < 1:
        return None
    output = [output[i] for i in has_pct]
    values = [_get_values(c) for c in output]
    values = pd.DataFrame(values, columns=["Conditions", "Model"])
    values["Variable"] = [_get_variable(c) for c in output]

    if values.shape[0] < x.shape[1]:
        x_names = set(x.columns)
        u_names = set(values["Variable"]) if values is not None else set()
        missing_vars = list(x_names - u_names)
        if missing_vars:
            zero_list = [0] * len(missing_vars)
            usage2 = pd.DataFrame({"Conditions": zero_list,
                                   "Model": zero_list,
                                   "Variable": missing_vars})
            values = pd.concat([values, usage2], axis=1)
            values = values.reset_index(drop=True)
    return values


def _get_values(x):
    x2 = x.split(" ")
    has_pct = ["%" in c for c in x2]
    if sum(has_pct) == 2:
        x2 = [x2[i] for i, c in enumerate(has_pct) if c]
        x2 = [float(c.replace("%", "")) for c in x2]
        return x2
    if sum(has_pct) == 1:
        pct_ind = [i for i, c in enumerate(x2) if "%" in c][0]
        if x2[1:pct_ind+1].count("") < x2[pct_ind+1:].count(""):
            x2 = [c for c in x2 if "%" in c][0]
            x2 = float(x2.replace("%", ""))
            return [x2, 0]
        x2 = [c for c in x2 if "%" in c][0]
        x2 = float(x2.replace("%", ""))
        return [0, x2]
    return 0


def _get_variable(x):
    return x.split(" ")[-1]



import re
import sys
from datetime import datetime

from ._quinlan_attributes import _quinlan_attributes


def _make_names_string(x, w=None, label="outcome"):
    """
    Create the names string to pass to Cubist. This string contains information about Python and the time of run along
    with the column names and their data types.

    Parameters
    ----------
    x : {pd.DataFrame} of shape (n_samples, n_features)
        The input samples.

    w : ndarray of shape (n_samples,)
        Instance weights.

    label : str, default="outcome"
        A label for the outcome variable. This is only used for printing rules.

    Returns
    -------
    out : str
        Case name string describing training dataset columns and their types.
    """
    # copy Pandas objects so they aren't changed outside of this function
    x = x.copy(deep=True)

    # clean reserved sample name if it's in x
    has_sample = [i for i, c in enumerate(x.columns) if bool(re.search("^sample", c))]
    if has_sample:
        x.columns = [re.sub("^sample", "_Sample", c) for c in x.columns]

    # generate the comments string showing the Python version and current timestamps
    python_version = tuple(sys.version_info)
    now = datetime.now()
    out = (
        f'| Generated using Python {python_version[0]}.{python_version[1]}.{python_version[2]}\n'
        f'| on {now.strftime("%a %b %d %H:%M:%S %Y")}'
    )

    # define the outcome data type
    outcome_type = ": continuous."

    # build base out string
    out = f"{out}\n{label}.\n{label}{outcome_type}"

    # get dictionary of feature names as keys and data types as values
    var_data = _quinlan_attributes(x)

    # if weights are present add this to var_data
    if w is not None:
        var_data["case weight"] = "continuous."

    # join the column names and data types into a single string
    var_data = [f"{_escapes([key])[0]}: {value}" for key, value in var_data.items()]
    var_data = "\n".join(var_data)

    # merge the out and var_data strings
    out = f"{out}\n{var_data}\n"
    return out


def _escapes(x):
    """Double escape reserved and special characters in x."""
    # set custom reserved characters list
    chars = [":", ";", "|"]
    # apply first escaping
    for i in chars:
        x = [c.replace(i, f"\\{i}") for c in x]
    # apply second escaping
    return [_re_escape(c) for c in x]


_special_chars_map = {i: "\\" + chr(i) for i in b"()[]{}?*+-|:;^$\\.&~#\t\n\r\v\f"}


def _re_escape(pattern):
    """Escape special characters in a string.
    Sourced from 're' Python package."""
    if isinstance(pattern, str):
        return pattern.translate(_special_chars_map)
    pattern = str(pattern, "latin1")
    return pattern.translate(_special_chars_map).encode("latin1")



"""
cubist: Cubist regression model for Python
==========================================

cubist is a Python module that wraps the C implementation of Quinlan's
regression model of the same name. In the module, "Cubist" is the name of the
model class.

The model requires input data to be particularly formattted as strings and this
module handles this formatting process as well as the model options.

The module is designed after the sklearn API and inherits from the
RegressorMixin and BaseEstimator classes, meaning that Cubist can be dropped
into existing sklearn-based ML pipelines or scripts for experimenting.
"""
from .cubist import Cubist
from .exceptions import CubistError

__all__ = ["Cubist", "CubistError"]



import pandas as pd
import numpy as np


def _format(x: float, digits: int = 15) -> str:
    """Python version of the R format function to return a number formatted as a
    string rounded to `digits` number of digits from the left."""
    # if x is NA return NA
    if pd.isna(x):
        return x
    if np.iscomplex(x):
        raise ValueError("Complex numbers not supported")

    return np.format_float_positional(
        x, precision=digits, unique=False, fractional=False, trim="-"
    )



cimport numpy as np
np.import_array()

# external declarations for cubist and predictions function from the top.c file
cdef extern from "src/top.c":
    void cubist(char **namesv, char **datav, int *unbiased,
                char **compositev, int *neighbors, int *committees,
                double *sample, int *seed, int *rules, double *extrapolation,
                int *cv, char **modelv, char **outputv)
    void predictions(char **casev, char **namesv, char **datav, char **modelv, 
                     double *predv, char **outputv)

# define the Python functions that interface with the C functions
def _cubist(namesv_, datav_, unbiased_, compositev_, neighbors_, committees_, 
            sample_, seed_, rules_, extrapolation_, cv_, modelv_, outputv_):
    """
    Train and return Cubist model and output from C code
    """
    cdef char *namesv = namesv_;
    cdef char *datav = datav_;
    cdef int unbiased = unbiased_;
    cdef char *compositev = compositev_;
    cdef int neighbors = neighbors_;
    cdef int committees = committees_;
    cdef double sample = sample_;
    cdef int seed = seed_;
    cdef int rules = rules_;
    cdef double extrapolation = extrapolation_;
    cdef int cv = cv_;
    cdef char *modelv = modelv_;
    cdef char *outputv = outputv_;
    cubist(&namesv, &datav, &unbiased, &compositev, &neighbors, &committees, 
           &sample, &seed, &rules, &extrapolation, &cv, &modelv, &outputv);
    return (modelv, outputv)


def _predictions(casev_, namesv_, datav_, modelv_, 
                 np.ndarray[double, ndim=1, mode="c"] predv_, outputv_):
    """
    Obtain predictions using existing Cubist model and return output if raised
    Reference: https://scipy-lectures.org/advanced/interfacing_with_c/interfacing_with_c.html#id13
    """
    cdef char *casev = casev_;
    cdef char *namesv = namesv_;
    cdef char *datav = datav_;
    cdef char *modelv = modelv_;
    cdef char *outputv = outputv_;
    predictions(&casev, &namesv, &datav, &modelv, 
                <double*> np.PyArray_DATA(predv_), &outputv);
    return (predv_, outputv)



import pandas as pd
from pandas.api.types import is_string_dtype, is_numeric_dtype, \
    is_datetime64_any_dtype, is_complex_dtype
import numpy as np


def _is_all_float_dtype(x: pd.Series):
    """check whether all values are of float dtype"""
    return all(j == float or np.issubdtype(j, np.floating) for j in [type(i) for i in x.values])


def _is_all_int_dtype(x: pd.Series):
    """check whether all values are of float dtype"""
    return all(j == int or np.issubdtype(j, np.integer) for j in [type(i) for i in x.values])


def _get_data_format(x: pd.Series):
    """
    Function to obtain the data type/formatting information for a Pandas Series.
    Return "continuous." for continuous features, the set of values as a comma
    separated string for categorical features, and the column itself for 
    datetime features.

    Parameters
    ----------
    x : pd.Series
        Pandas Series from which to extract data type.
    
    Returns
    -------
    x : str
        String description of the Series data type.
    """
    # remove NAs from series
    x = x.dropna()
    if is_complex_dtype(x):
        raise ValueError("Complex data not supported")
    # for numeric columns
    if is_numeric_dtype(x) or _is_all_float_dtype(x) or _is_all_int_dtype(x):
        return "continuous."
    # for string columns
    if is_string_dtype(x):
        x = x.astype(str)
        return f"{','.join(set(x))}."
    # for datetime columns
    if is_datetime64_any_dtype(x):
        return x
    raise ValueError(f"Dtype {x.dtype} is not supported")


def _quinlan_attributes(df: pd.DataFrame) -> dict:
    """
    Function to collect the data formatting information for each column in a Pandas DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame from which column data attributes are obtained.
    
    Returns
    -------
    x : dict
        Dictionary with keys as column names and values as the description of
        the data type.
    """
    return {col_name: _get_data_format(df[col_name]) for col_name in df.columns}



import pandas as pd
from pandas.api.types import is_string_dtype, is_numeric_dtype
import numpy as np

from ._make_names_string import _escapes
from ._utils import _format


def _make_data_string(x, y=None, w=None):
    """
    Converts input dataset array X into a string.

    Parameters
    ----------
    x : {pd.DataFrame} of shape (n_samples, n_features)
        The input samples.

    y : pd.Series
        The predicted values.

    w : ndarray of shape (n_samples,)
        Instance weights.

    Returns
    -------
    x : str
        Input dataset converted to a string and formatted per Cubist's
        requirements.
    """
    x = x.copy(deep=True)

    # apply the escapes function to all string columns
    for col in x:
        if is_string_dtype(x[col]):
            x[col] = _escapes(x[col].astype(str))

    # if y is None for model predictions, set y as a column of NaN values,
    # # which will become ?'s later
    if y is None:
        y = [np.nan] * x.shape[0]
        y = pd.Series(y)
    else:
        y = y.copy(deep=True)

    # format the y column for special charactesrs
    y = pd.Series(_escapes(y.astype(str)))

    # insert the y column as the first column of x
    x.insert(0, "y", y)

    # handle weights matrix (?) TODO: validate
    if w is not None:
        # [f"w{i}" for i in range(x.shape[1])]
        column_names = list(x.columns) + ["w"]
        x = x.assign(w=w)
        x.columns = column_names

    # convert all columns to strings
    for col in x:
        if is_numeric_dtype(x[col]):
            x[col] = x[col].apply(_format)
            x[col] = x[col].astype(str)
        else:
            x[col] = x[col].astype(str)

    # remove leading whitespace from all elements
    x = x.applymap(lambda a: a.lstrip())

    # replace missing values with ?
    x = x.fillna("?")
    x = x.replace("nan", "?")

    # convert dataframe to list of lists
    x = x.to_numpy().tolist()

    # merge each sublist into single strings with entries separated by commas
    x = [",".join(row) for row in x]

    # join all row strings into a single string separated by \n's
    x = "\n".join(x)
    return x



import re
import math
import operator

import pandas as pd
import numpy as np

from ._utils import _format


OPERATORS = {
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
    "==": operator.eq,
}


def _split_to_groups(x, f):
    """Function to convert two lists into a dictionary where the keys are
    unique values in f and the values are lists of the corresponding values in
    x. Analogous to the split function in R."""
    if len(x) != len(f):
        raise ValueError("lists x and f must be of the same length")
    groups = {}
    for a, b in zip(x, f):
        if b in groups:
            groups[b].append(a)
        else:
            groups[b] = [a]
    return groups


def _parse_model(model, x):
    # split on newline
    model = model.split("\n")
    # remove empty strings
    model = [c for c in model if c.strip() != ""]
    # get model length
    model_len = len(model)

    # define initial lists and index variables
    com_num = [None] * model_len
    rule_num = [None] * model_len
    cond_num = [None] * model_len
    com_idx = r_idx = 0

    # loop through model and indicate
    for i, row in enumerate(model):
        # break each row of x into dicts for each key/value pair
        tt = _parser(row)
        # get the first key in the first entry of tt
        first_key = list(tt[0])[0]

        # start of a new committee and rule
        if first_key == "rules":
            # increment committee number
            com_idx += 1
            # reset rule number
            r_idx = 0
        # set the current committee number
        com_num[i] = com_idx
        # start of a new rule
        if first_key == "conds":
            # increment rule number
            r_idx += 1
            # reset condition number
            c_idx = 0
        # set the current rule number
        rule_num[i] = r_idx
        # within a rule, type designates the type of conditional statement
        # type = 2 appears to be a simple split of a continuous predictor
        if first_key == "type":
            # increment condition number
            c_idx += 1
            # set the current condition number
            cond_num[i] = c_idx

    split_var = [None] * model_len
    split_val = [None] * model_len
    split_cats = [""] * model_len
    split_dir = [""] * model_len
    split_type = [""] * model_len

    # handle continuous (type 2) rules
    is_type2 = [i for i, c in enumerate(model) if re.search('^type="2"', c)]
    for i in is_type2:
        # set the type of split
        split_type[i] = "continuous"
        continuous_split = _type2(model[i])
        # set split variable name
        split_var[i] = continuous_split["var"].replace('"', "")
        # set split direction (comparison operator)
        split_dir[i] = continuous_split["result"]
        # set split value
        split_val[i] = continuous_split["val"]

    # handle categorical (type 3) rules
    is_type3 = [i for i, c in enumerate(model) if re.search('^type="3"', c)]
    for i in is_type3:
        categorical_split = _type3(model[i])
        # set the type of split
        split_type[i] = "categorical"
        # set the split variable
        split_var[i] = categorical_split["var"]
        # set the split value
        split_cats[i] = categorical_split["val"]

    # if there are no continuous or categorical splits, return no splits
    if is_type2 == [] and is_type3 == []:
        split_data = None
    else:
        split_data = pd.DataFrame(
            {
                "committee": com_num,
                "rule": rule_num,
                "variable": split_var,
                "dir": split_dir,
                "value": split_val,
                "category": split_cats,
                "type": split_type,
            }
        )
        # remove missing values based on the variable column
        split_data = split_data.dropna(subset=["variable"])
        split_data = split_data.reset_index(drop=True)

        # get the percentage of data covered by this rule
        nrows = x.shape[0]
        for i in range(split_data.shape[0]):
            # get the current value threshold and comparison operator
            var_value = split_data.loc[i, "value"]
            comp_operator = split_data.loc[i, "dir"]
            if var_value is not None:
                if not math.isnan(var_value):
                    # convert the data to numeric and remove NaNs
                    x_col = pd.to_numeric(x[split_data.loc[i, "variable"]]).dropna()
                    # evaluate and get the percentage of data
                    comp_total = OPERATORS[comp_operator](x_col, var_value).sum()
                    split_data.loc[i, "percentile"] = comp_total / nrows

    # get the indices of rows in model that contain model coefficients
    is_eqn = [i for i, c in enumerate(model) if "coeff=" in c]
    # extract the model coefficients from the row
    coefs = [_eqn(model[i], var_names=list(x.columns)) for i in is_eqn]
    out = pd.DataFrame(coefs)
    # get the committee number
    out["committee"] = [com_num[i] for i in is_eqn]
    # get the rule number for the committee
    out["rule"] = [rule_num[i] for i in is_eqn]

    return split_data, out


def _type2(x, dig=3):
    x = x.replace('"', "")

    # get the indices where these keywords start
    att_ind = x.find("att=")
    cut_ind = x.find("cut=")
    result_ind = x.find("result=")
    val_ind = x.find("val=")

    missing_rule = cut_ind < 1 and val_ind > 0
    if missing_rule:
        var = x[att_ind + 4 : cut_ind - 1]
        val = None
        result = "="
    else:
        var = x[att_ind + 4 : cut_ind - 1]
        val = x[cut_ind + 4 : result_ind - 1]
        val = _format(float(val), dig)
        result = x[result_ind + 7 :]
    return {
        "var": var,
        "val": float(val),
        "result": result,
        "text": f"{var} {result} {val}",
    }


def _type3(x):
    # get the indices where these keywords start
    att_ind = x.find("att=")
    elts_ind = x.find("elts=")

    var = x[att_ind + 5 : elts_ind - 2]
    val = x[elts_ind + 5 :]
    val = val.replace("[{}]", "").replace('"', "").replace(" ", "").replace(",", ", ")

    # if there are multiple values in the categorical split, just show
    # {multiple_vals} for cleaner printing}
    # TODO: enter all vals in dataframe but limit the column width when printing
    multiple_vals = "," in val
    if multiple_vals:
        val = "{" + str(multiple_vals) + "}"
        txt = f"{var} in {val}"
    else:
        txt = f"{var} = {val}"
    return {"var": var, "val": val, "text": txt}


def _eqn(x, var_names=None):
    x = x.replace('"', "")
    starts = [m.start(0) for m in re.finditer("(coeff=)|(att=)", x)]
    tmp = [""] * len(starts)
    for i, val in enumerate(starts):
        if i < len(starts) - 1:
            txt = x[val : starts[i + 1] - 1]
        else:
            txt = x[val:]
        tmp[i] = txt.replace("coeff=", "").replace("att=", "")

    vals = tmp[::2]
    vals = [float(c) for c in vals]
    nms = tmp[1::2]
    nms = ["(Intercept)"] + nms
    vals = dict(zip(nms, vals))
    if var_names:
        vars2 = [var for var in var_names if var not in nms]
        vals2 = [np.nan] * len(vars2)
        vals2 = dict(zip(vars2, vals2))
        vals = {**vals, **vals2}
        new_names = ["(Intercept)"] + var_names
        vals = {nm: vals[nm] for nm in new_names}
    return vals


def _make_parsed_dict(x):
    x = x.split("=")
    if len(x) > 1:
        return {x[0]: x[1]}
    return None


def _parser(x):
    x = x.split(" ")
    x = [_make_parsed_dict(c) for c in x]
    return x



import zlib
from warnings import warn

import numpy as np
import pandas as pd
from sklearn.utils.validation import (
    check_is_fitted,
    check_random_state,
    _check_sample_weight,
)
from sklearn.base import RegressorMixin, BaseEstimator

from _cubist import _cubist, _predictions  # noqa E0611 # pylint: disable=E0611

from ._make_names_string import _make_names_string
from ._make_data_string import _make_data_string
from ._parse_model import _parse_model
from ._variable_usage import _get_variable_usage
from .exceptions import CubistError


class Cubist(BaseEstimator, RegressorMixin):
    """
    Cubist Regression Model (Public v2.07) developed by Quinlan.

    References:
    - https://www.rdocumentation.org/packages/Cubist/versions/0.3.0
    - https://www.rulequest.com/cubist-unix.html

    Parameters
    ----------
    n_rules : int, default=500
        Limit of the number of rules Cubist will build. Recommended and default
        value is 500.

    n_committees : int, default=1
        Number of committees to construct. Each committee is a rule based model
        and beyond the first tries to correct the prediction errors of the prior
        constructed model. Recommended value is 5.

    neighbors : int, default=None
        Number between 1 and 9 for how many instances should be used to correct
        the rule-based prediction. If no value is given, Cubist will build a
        rule-based model only. If this value is set, Cubist will create a
        composite model with the given number of neighbors. Regardless of the
        value set, if auto=True, Cubist may override this input and choose a
        different number of neighbors. Please assess the model for the selected
        value for the number of neighbors used.

    unbiased : bool, default=False
        Should unbiased rules be used? Since Cubist minimizes the MAE of the
        predicted values, the rules may be biased and the mean predicted value
        may differ from the actual mean. This is recommended when there are
        frequent occurrences of the same value in a training dataset. Note that
        MAE may be slightly higher.

    auto : bool, default=False
        A value of True allows the algorithm to choose whether to use
        nearest-neighbor corrections and how many neighbors. In this case,
        `neighbors` must be left as None since the input has no bearing on the
        model's behavior. False will leave the choice of whether to use a
        composite model to value passed to the `neighbors` parameter.

    extrapolation : float, default=0.05
        Adjusts how much rule predictions are adjusted to be consistent with
        the training dataset. Recommended value is 5% as a decimal (0.05)

    sample : float, default=None
        Percentage of the data set to be randomly selected for model building.

    cv : int, default=None
        Whether to carry out cross-validation and how many folds to use
        (recommended value is 10 per Quinlan)

    random_state : int, default=None
        An integer to set the random seed for the C Cubist code.

    target_label : str, default="outcome"
        A label for the outcome variable. This is only used for printing rules.

    verbose : int, default=0
        Should the Cubist output be printed?

    Attributes
    ----------
    names_string_ : str
        String for the Cubist model that describes the training dataset column
        names and their data types. This also provides some Python environment
        information.

    data_string_ : str
        String containing the training data. Required for using instance-based
        corrections and compressed after model training.

    model_ : str
        The Cubist model string generated by the C code.

    feature_importances_ : pd.DataFrame
        Table of how training data variables are used in the Cubist model. The
        first column for "Conditions" shows the approximate percentage of cases
        for which the named attribute appears in a condition of an applicable
        rule, while the second column "Attributes" gives the percentage of cases
        for which the attribute appears in the linear formula of an applicable
        rule.

    rules_ : pd.DataFrame
        Table of the rules built by the Cubist model.

    coeff_ : pd.DataFrame
        Table of the regression coefficients found by the Cubist model.

    variables_ : dict
        Information about all the variables passed to the model and those that
        were actually used.

    Examples
    --------
    >>> from cubist import Cubist
    >>> from sklearn.datasets import fetch_california_housing
    >>> from sklearn.model_selection import train_test_split
    >>> X, y = fetch_california_housing(return_X_y=True, as_frame=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y,
                                                            test_size=0.2,
                                                            random_state=42)
    >>> model = Cubist()
    >>> model.fit(X_train, y_train)
    >>> model.predict(X_test)
    >>> model.score(X_test, y_test)
    """

    def __init__(
        self,
        n_rules: int = 500,
        *,
        n_committees: int = 1,
        neighbors: int = None,
        unbiased: bool = False,
        auto: bool = False,
        extrapolation: float = 0.05,
        sample: float = None,
        cv: int = None,
        random_state: int = None,
        target_label: str = "outcome",
        verbose: int = 0,
    ):
        super().__init__()

        self.n_rules = n_rules
        self.n_committees = n_committees
        self.neighbors = neighbors
        self.unbiased = unbiased
        self.auto = auto
        self.extrapolation = extrapolation
        self.sample = sample
        self.cv = cv
        self.random_state = random_state
        self.target_label = target_label
        self.verbose = verbose

    def _more_tags(self):
        """scikit-learn estimator configuration method"""
        return {"allow_nan": True, "X_types": ["2darray", "string"]}

    def _check_n_rules(self):
        # validate number of rules
        if not isinstance(self.n_rules, int):
            raise TypeError(
                f"`n_rules` must be an integer but got {type(self.n_rules)}"
            )
        if self.n_rules < 1 or self.n_rules > 1000000:
            raise ValueError(
                f"`n_rules` must be between 1 and 1000000 but got {self.n_rules}"
            )
        return self.n_rules

    def _check_n_committees(self):
        # validate number of committees
        if not isinstance(self.n_committees, int):
            raise TypeError(
                f"`n_committees` must be an integer but got {type(self.n_committees)}"
            )
        if self.n_committees < 1 or self.n_committees > 100:
            raise ValueError(
                f"`n_committees` must be between 1 and 100 but got {self.n_committees}"
            )
        return self.n_committees

    def _check_neighbors(self):
        # if auto is True but neighbors is set, raise an error
        if self.auto:
            if self.neighbors is not None:
                raise ValueError(
                    f"When `auto`=True, `neighbors` should be None as Cubist will choose this parameter but got {self.neighbors}."
                )
            return 0
        # default value must be zero even when not used
        if self.neighbors is None:
            return 0
        # validate number of neighbors
        if not isinstance(self.neighbors, int):
            raise TypeError(
                f"`neighbors` must be an integer but got {type(self.neighbors)}"
            )
        if self.neighbors < 1 or self.neighbors > 9:
            raise ValueError(
                f"`neighbors` must be between 1 and 9 but got {self.neighbors}"
            )
        return self.neighbors

    def _check_unbiased(self):
        # validate unbiased option
        if not isinstance(self.unbiased, bool):
            raise TypeError(
                f"`unbiased` must be a boolean but got {type(self.unbiased)}"
            )
        return self.unbiased

    def _check_composite(self, neighbors):
        # validate the auto parameter
        if not isinstance(self.auto, bool):
            raise TypeError(f"`auto` must be a boolean but got {type(self.auto)}")
        # if auto=True, let cubist decide whether to use a composite model and
        # how many neighbors to use
        if self.auto:
            return "auto"
        # if a number of neighbors is given, make a composite model
        if neighbors > 0:
            return "yes"
        return "no"

    def _check_extrapolation(self):
        # validate the range of extrapolation
        if not isinstance(self.extrapolation, float):
            raise TypeError(
                f"`extrapolation` must be a float but got {type(self.extrapolation)}"
            )
        if self.extrapolation < 0.0 or self.extrapolation > 1.0:
            raise ValueError(
                f"`extrapolation` must be between 0.0 and 1.0 but got {self.extrapolation}"
            )
        return self.extrapolation

    def _check_sample(self, num_samples):
        # default value must be 0 if not used
        if self.sample is None:
            return 0
        # validate the sample type
        if not isinstance(self.sample, float):
            raise TypeError(f"`sample` must be a float but got {type(self.sample)}")
        # validate sample value
        if not (0.0 < self.sample < 1.0):
            raise ValueError(
                f"`sample` must be between 0.0 and 1.0 but got {self.sample}"
            )
        # check to see if the sample will create a very small dataset
        trained_num_samples = int(round(self.sample * num_samples, 0))
        if trained_num_samples < 10:
            warn(
                f"Sampling a dataset with {num_samples} rows and a "
                f"sampling percent of {self.sample} means Cubist will "
                f"train with {trained_num_samples} rows. This may lead "
                f"to incorrect or failing predictions. Please increase "
                f"or remove the `sample` parameter.\n",
                stacklevel=3,
            )
        return self.sample

    def _check_cv(self):
        # default value must be 0 if not used
        if self.cv is None:
            return 0
        # validate type
        if not isinstance(self.cv, int):
            raise TypeError(f"`cv` must be an integer but got {type(self.cv)}")
        # validate value
        if self.cv <= 1:
            raise ValueError(f"`cv` must be greater than 1 but got {self.cv}")
        return self.cv

    def fit(self, X, y, sample_weight=None):
        """Build a Cubist regression model from training set (X, y).

        Parameters
        ----------
        X : {array-like} of shape (n_samples, n_features)
            The training input samples. Must have complete column names or none
            provided at all (NumPy arrays will be given names by column index).

        y : array-like of shape (n_samples,)
            The target values (Real numbers in regression).

        sample_weight : array-like of shape (n_samples,)
            Optional vector of sample weights that is the same length as y for
            how much each instance should contribute to the model fit.

        Returns
        -------
        self : object
        """
        # scikit-learn data validation
        X, y = self._validate_data(
            X,
            y,
            dtype=None,
            force_all_finite="allow-nan",
            y_numeric=True,
            ensure_min_samples=2,
        )

        # set the feature names if it hasn't already been done
        if not hasattr(self, "feature_names_in_"):
            self.feature_names_in_ = [f"var{i}" for i in range(X.shape[1])]  # noqa W0201

        # check to see if any of the feature names are empty
        if any(n == "" or pd.isnull(n) for n in self.feature_names_in_):
            raise ValueError(
                "At least one column is missing unnamed and is NA \
                             or an empty string."
            )

        # check sample weighting
        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X)
            self.is_sample_weighted_ = True  # noqa W0201
        else:
            self.is_sample_weighted_ = False  # noqa W0201

        n_rules = self._check_n_rules()
        n_committees = self._check_n_committees()
        neighbors = self._check_neighbors()
        unbiased = self._check_unbiased()
        composite = self._check_composite(neighbors)
        extrapolation = self._check_extrapolation()
        sample = self._check_sample(X.shape[0])
        cv = self._check_cv()
        random_state = check_random_state(self.random_state)

        # number of input features
        self.n_features_in_ = X.shape[1]  # noqa W0201
        # number of outputs is 1 (single output regression)
        self.n_outputs_ = 1  # noqa W0201

        # (re)construct a dataframe from X
        X = pd.DataFrame(X, columns=self.feature_names_in_)
        y = pd.Series(y)

        # create the names and data strings required for cubist
        names_string = _make_names_string(X, w=sample_weight, label=self.target_label)
        data_string = _make_data_string(X, y, w=sample_weight)

        # call the C implementation of cubist
        model, output = _cubist(
            namesv_=names_string.encode(),
            datav_=data_string.encode(),
            unbiased_=unbiased,
            compositev_=composite.encode(),
            neighbors_=neighbors,
            committees_=n_committees,
            sample_=sample,
            seed_=random_state.randint(0, 4095) % 4096,
            rules_=n_rules,
            extrapolation_=extrapolation,
            cv_=cv,
            modelv_=b"1",
            outputv_=b"1",
        )

        # convert output from raw to strings
        self.model_ = model.decode()
        output = output.decode()

        # raise Cubist training errors
        if ("***" in output) or ("Error" in output):
            raise CubistError(output)

        # inform user that they may want to use rules only
        if "Recommend using rules only" in output:
            warn(
                "Cubist recommends using rules only (set auto=False)",
                stacklevel=3,
            )

        # print model output if using verbose output
        if self.verbose:
            print(output)

        # if the model returned nothing, we're doing cross-validation so stop
        if self.model_ == "1":
            return self

        # replace "__Sample" with "sample" if this is used in the model
        if "\n_Sample" in names_string:
            output = output.replace("_Sample", "sample")
            self.model_ = self.model_.replace("_Sample", "sample")
            # clean model string when using reserved sample name
            self.model_ = (
                self.model_[: self.model_.index("sample")]
                + self.model_[self.model_.index("entries") :]
            )

        # when a composite model has not been used, drop the data_string
        if not (
            (composite == "yes") or ("nearest neighbors" in output) or (neighbors > 0)
        ):
            data_string = "1"

        # compress and save descriptors/data
        self.names_string_ = zlib.compress(names_string.encode())  # noqa W0201
        self.data_string_ = zlib.compress(data_string.encode())  # noqa W0201

        # parse model contents and store useful information
        self.rules_, self.coeff_ = _parse_model(self.model_, X)  # noqa W0201

        # get the input data variable usage
        self.feature_importances_ = _get_variable_usage(output, X)  # noqa W0201

        # get the names of columns that have no nan values
        is_na_col = ~self.coeff_.isna().any()
        not_na_cols = self.coeff_.columns[is_na_col].tolist()

        # skip the first three since these are always filled
        not_na_cols = not_na_cols[3:]

        # store a dictionary containing all the training dataset columns and
        # those that were used by the model
        if self.rules_ is not None:
            used_variables = set(self.rules_["variable"]).union(set(not_na_cols))
            self.variables_ = {
                "all": list(self.feature_names_in_),
                "used": list(used_variables),
            }
        return self

    def predict(self, X):
        """Predict Cubist regression target for X.

        Parameters
        ----------
        X : {array-like} of shape (n_samples, n_features)
            The input samples. Must have complete column names or none
            provided at all (NumPy arrays will be given names by column index).

        Returns
        -------
        y : ndarray of shape (n_samples,)
            The predicted values.
        """
        # make sure the model has been fitted
        check_is_fitted(
            self, attributes=["model_", "rules_", "coeff_", "feature_importances_"]
        )

        # validate input data
        X = self._validate_data(
            X, dtype=None, force_all_finite="allow-nan", reset=False
        )

        # (re)construct a dataframe from X
        X = pd.DataFrame(X, columns=self.feature_names_in_)

        # If there are case weights used during training, the C code will expect
        # a column of weights in the new data but the values will be ignored.
        if self.is_sample_weighted_:
            X["case_weight_pred"] = np.nan

        # make data string for predictions
        data_string = _make_data_string(X)

        # get cubist predictions from trained model
        pred, output = _predictions(
            data_string.encode(),
            zlib.decompress(self.names_string_),
            zlib.decompress(self.data_string_),
            self.model_.encode(),
            np.zeros(X.shape[0]),
            b"1",
        )

        # decode output
        output = output.decode()

        # raise Cubist prediction errors
        if "***" in output or "Error" in output:
            raise CubistError(output)

        if output:
            print(output)

        return pred



import glob

from Cython.Build import cythonize
from setuptools import setup, Extension
import numpy as np


extensions = [
    Extension(
        name="_cubist",
        sources=["cubist/_cubist.pyx"] + glob.glob("cubist/src/*.c"),
        include_dirs=["cubist/src", np.get_include()],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    )
]

setup(
    ext_modules=cythonize(
        extensions, compiler_directives={"language_level": 3, "profile": False}
    ),
)
